#!/usr/bin/env python
from __future__ import with_statement, division, print_function
import operator, sys, itertools
from array import array
import pb53


infilename = "SolarWars2001.NES"

# Other patches in the same part of the ROM as the CNROM to UNROM
# patches have to be listed here so they can get inserted in the
# proper order.
# These fix a stack overflow per 43110 in
# http://forums.nesdev.com/viewtopic.php?p=136686#p136686
other_patches = [
    (0xDB8C, "jmp $dba9  ; stack overflow fix"),
    (0xDC4C, "jmp $dc69  ; stack overflow fix"),
]


def icum(iterable, func=operator.add):
    """Return running totals.

Equivalent to accumulate() in Python 3.3 or later.

list(icum([8,6,7,5,3,0,9])) == [8,14,21,26,29,29,38]
list(icum([8,6,7,5,3,0,9], operator.mul)) == [8,48,336,1680,5040,0,0]
list(icum([8,6,7,5,3,0,9], min)) == [8, 6, 6, 5, 3, 0, 0]

For only the last value, use reduce().

"""
    it = iter(iterable)
    total = next(it)
    yield total
    for el in it:
        total = func(total, el)
        yield total


def find_c2c0(prgbytes, print_matches=False):
    """Find writes to the bus conflict avoidance table at $C2C0."""
    offsets = [i for i in xrange(len(prgbytes) - 3)
               if (prgbytes[i + 2] == 0xC2
                   and 0xC0 <= prgbytes[i + 1] < 0xC4)]
    if print_matches:
        print("\n".join("%04X: %s"
                        % (0x8000 + i,
                           ' '.join("%02X" % b for b in prgbytes[i:i + 3]))
                        for i in offsets))

    # Turns all useful ones are 9D C0 C2 = STA $C2C0,X
    # provided it doesn't write through (d,x) or (d),y mode
    offsets = [i for i in offsets
               if prgbytes[i] == 0x9D and prgbytes[i + 1] == 0xC0]
    return offsets

with open(infilename, 'rb') as infp:
    inesheader = infp.read(16)
    prgrom = infp.read(32768)
    chrbanks = [infp.read(8192) for x in xrange(4)]

# Load the banks
prgbytes = array('B', prgrom)
patch_offsets = find_c2c0(prgbytes)
print("Writing patch offsets to loadbank_patches.s")
all_patches = [(0x8000 + x, "jsr loadbank  ; CNROM to UNROM")
               for x in patch_offsets]
all_patches.extend(other_patches)
all_patches.sort()
print("\n".join(repr(row) for row in all_patches))
with open("loadbank_patches.s", "wt") as outfp:
    outfp.writelines("  copy_till $%04x\n  %s\n" % patch
                     for patch in all_patches)

# Compress the banks
pb53banks = [pb53.pb53(chrdata, segsize=8192) for chrdata in chrbanks]
lengths = [len(pb53data) for pb53data, splits in pb53banks]
print("Writing PB53 data to chrbank#.bin")
print("Lengths: %s, total %d" % (repr(lengths), sum(lengths)))
for i, (pb53data, _) in enumerate(pb53banks):
    with open("chrbank%d.bin" % i, "wb") as outfp:
        outfp.write(pb53data)
